/*
 * @copyright Copyright (c) OX Software GmbH, Germany <info@open-xchange.com>
 * @license AGPL-3.0
 *
 * This code is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with OX App Suite.  If not, see <https://www.gnu.org/licenses/agpl-3.0.txt>.
 *
 * Any use of the work other than as authorized under this license or copyright law is prohibited.
 *
 */

package com.openexchange.util.activator.impl;

import static com.google.common.collect.ImmutableList.toImmutableList;
import static com.openexchange.util.custom.base.NullUtil.build;
import static com.openexchange.util.custom.base.NullUtil.immutableSetCopyOf;
import static com.openexchange.util.custom.base.NullUtil.listBuilder;
import static com.openexchange.util.custom.base.NullUtil.logger;
import static com.openexchange.util.custom.base.NullUtil.mapBuilder;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Dictionary;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.Set;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.concurrent.atomic.AtomicReference;
import java.util.concurrent.locks.ReentrantLock;
import javax.annotation.concurrent.ThreadSafe;
import org.osgi.framework.BundleContext;
import org.osgi.framework.ServiceReference;
import org.slf4j.Logger;
import com.google.common.collect.ImmutableList;
import com.google.common.collect.ImmutableMap;
import com.google.common.collect.ImmutableSet;
import com.openexchange.annotation.NonNull;
import com.openexchange.annotation.Nullable;
import com.openexchange.config.Reloadable;
import com.openexchange.osgi.DeferredActivator;
import com.openexchange.osgi.ServiceListing;
import com.openexchange.osgi.ServiceSet;
import com.openexchange.osgi.console.ServiceStateLookup;
import com.openexchange.util.activator.ServiceDependencyResolver;
import com.openexchange.util.custom.base.NullUtil;
import com.openexchange.util.custom.collect.ImmutableDictionary;

/**
 * This abstract service tracker can wait for some service and invoke {@link #register}
 * when all dependencies are available.
 *
 * @author <a href="mailto:pascal.bleser@open-xchange.com">Pascal Bleser</a>
 * @author <a href="mailto:marcus.klein@open-xchange.com">Marcus Klein</a>
 */
@ThreadSafe
public abstract class ServiceAwareDependentServiceRegistererTemplate<S, T extends S> implements ExtendedServiceTrackerCustomizer<S, T> {

    private static final Logger LOG = logger(ServiceAwareDependentServiceRegistererTemplate.class);
    @SuppressWarnings("null")
    private static final String NL = String.format("%n");
    
    private static final class ResolvedService {
        public final Object serviceObject; 
        public final Dictionary<String, ?> properties;
        public ResolvedService(Object serviceObject, Dictionary<String, ?> properties) {
            this.serviceObject = serviceObject;
            this.properties = properties;
        }
        @Override
        public int hashCode() {
            return Objects.hash(serviceObject, properties);
        }
        @Override
        public boolean equals(@Nullable Object obj) {
            if (obj == null) {
                return false;
            }
            if (! obj.getClass().equals(this.getClass())) {
                return false;
            }
            final ResolvedService other = (ResolvedService) obj;
            return Objects.equals(serviceObject, other.serviceObject)
                && Objects.equals(properties, other.properties);
        }
    }
    
    private final ReentrantLock lock = new ReentrantLock();
    protected final BundleContext context;
    protected final ImmutableSet<Class<?>> serviceTypes;
    protected final String serviceInstanceName;
    protected final Dictionary<String, ?> properties;
    protected final ImmutableMap<Class<?>, ServiceSet<?>> serviceSets;
    protected final ImmutableMap<Class<?>, ServiceListing<?>> serviceListings;
    protected final ImmutableSet<String> propertiesOfInterest;
    private final ImmutableMap<MandatoryServiceDependency<?>, AtomicReference<ResolvedService>> mandatoryDependencyMap;
    private final ImmutableList<OptionalServiceDependency<?>> optionalDependencyList;
    private final AtomicBoolean registered = new AtomicBoolean(false);

    public ServiceAwareDependentServiceRegistererTemplate(
        BundleContext context,
        ImmutableSet<Class<?>> serviceTypes,
        String serviceInstanceName,
        Dictionary<String, ?> properties,
        Collection<MandatoryServiceDependency<?>> mandatoryServices,
        Collection<OptionalServiceDependency<?>> optionalServices,
        ImmutableMap<Class<?>, ServiceSet<?>> serviceSets,
        ImmutableMap<Class<?>, ServiceListing<?>> serviceListings,
        Set<String> propertiesOfInterest) {
        super();
        this.context = context;
        this.serviceTypes = serviceTypes;
        this.serviceInstanceName = serviceInstanceName;
        this.properties = properties;
        {
            final ImmutableMap.Builder<MandatoryServiceDependency<?>, AtomicReference<ResolvedService>> mandatory = mapBuilder();
            for (final MandatoryServiceDependency<?> neededService : mandatoryServices) {
                if (neededService != null) {
                    mandatory.put((MandatoryServiceDependency<?>) neededService, new AtomicReference<ResolvedService>(null));
                }
            }
            this.mandatoryDependencyMap = build(mandatory);

            {
                @SuppressWarnings("null")
                final @NonNull ImmutableList<OptionalServiceDependency<?>> r = optionalServices.stream().filter(Objects::nonNull).collect(toImmutableList());
                this.optionalDependencyList = r;
            }
        }
        this.serviceSets = serviceSets;
        this.serviceListings = serviceListings;
        this.propertiesOfInterest = immutableSetCopyOf(propertiesOfInterest);
        setState();
    }
    
    @Override
    public final void onOpened() {
        try {
            if (mandatoryDependencyMap.isEmpty() && ! optionalDependencyList.isEmpty()) {
                if (registered.compareAndSet(false, true)) {
                    LOG.trace("{} has no mandatory dependencies, proceeding to registration rightaway", serviceInstanceName);
                    registerService(snapshotDependencies());
                }
            }
        } finally {
            onOpenedExtension();
        }
    }
    
    protected void onOpenedExtension() {
    }

    private static ImmutableDictionary<String, ?> snapshotProperties(final ServiceReference<?> reference, final Set<String> keys) {
        final ImmutableMap.Builder<String, Object> b = ImmutableMap.builder();
        for (final String key : keys) {
            final Object value = reference.getProperty(key);
            if (key != null && value != null) {
                b.put(key, value);
            }
        }
        return ImmutableDictionary.copyOf(b.build());
    }
    
    @Override
    public final T addingService(final @Nullable ServiceReference<S> reference) {
        if (reference == null) {
            throw new IllegalArgumentException("ServiceReference parameter is null");
        }
        final T obj;
        {
            final S s = context.getService(reference);
            if (s == null) {
                throw new IllegalStateException("reference resolution returned null");
            }
            @SuppressWarnings("unchecked") final T t = (T) s;
            obj = t;
        }
        final ResolvedService r = new ResolvedService(obj, snapshotProperties(reference, propertiesOfInterest));
        boolean needsRegistration = true;
        final List<ServiceDependency<?>> unresolvedMandatoryServiceDependencies = new LinkedList<>(); 
        lock.lock();
        try {
            for (final Map.Entry<MandatoryServiceDependency<?>, AtomicReference<ResolvedService>> e : mandatoryDependencyMap.entrySet()) {
                final MandatoryServiceDependency<?> dep = e.getKey();
                if (dep != null && dep.matchesService(obj, reference)) {
                    final ResolvedService oldService = e.getValue().getAndSet(r);
                    needsRegistration &= (oldService == null);
                    if (LOG.isTraceEnabled() && oldService == null) {
                        LOG.trace("dependency resolution for {}: inbound service '{}' matches dependency '{}'", serviceInstanceName, obj.getClass().getName(), dep.identify());
                    }
                } else {
                    if (e.getValue().get() == null) {
                        unresolvedMandatoryServiceDependencies.add(e.getKey());
                        if (LOG.isTraceEnabled()) {
                            if (dep != null) {
                                LOG.trace("dependency resolution for {}: still unresolved: '{}'", serviceInstanceName, dep.identify());
                            } else {
                                LOG.warn("dependency resolution for {}: service dependency is null", serviceInstanceName);
                            }
                        }
                    }
                }
            }
        } finally {
            lock.unlock();
        }
        if (! unresolvedMandatoryServiceDependencies.isEmpty()) {
            if (LOG.isTraceEnabled()) {
                final StringBuilder b = new StringBuilder();
                b.append(serviceInstanceName).append(" has ").append(unresolvedMandatoryServiceDependencies.size()).append(" unresolved dependencies:");
                for (final ServiceDependency<?> dependency : unresolvedMandatoryServiceDependencies) {
                    b.append(NL).append("- ").append(dependency.identify());
                }
                LOG.trace(b.toString());
            }
            needsRegistration = false;
        }
        
        for (final OptionalServiceDependency<?> dep : optionalDependencyList) {
            if (dep != null && dep.matchesService(obj, reference)) {
                dep.onResolved(obj);
            }
        }
        
        if (needsRegistration) {
            if (registered.compareAndSet(false, true)) {
                LOG.trace("{} has its {} mandatory service dependencies resolved, proceeding to registration",
                    serviceInstanceName, mandatoryDependencyMap.size());
                registerService(snapshotDependencies());
            }
        }
        setState();
        return obj;
    }
    
    private final ServiceDependencyResolver snapshotDependencies() {
        final ImmutableMap.Builder<MandatoryServiceDependency<?>, Object> m = mapBuilder();
        final ImmutableList.Builder<OptionalServiceDependency<?>> o = listBuilder();
        lock.lock();
        try {
            for (final Map.Entry<MandatoryServiceDependency<?>, AtomicReference<ResolvedService>> e : mandatoryDependencyMap.entrySet()) {
                final MandatoryServiceDependency<?> dep = e.getKey();
                if (dep != null) {
                    final ResolvedService value = e.getValue().get();
                    if (value != null) {
                        m.put(dep, value.serviceObject);
                    }
                }
            }
            for (final OptionalServiceDependency<?> dep : optionalDependencyList) {
                if (dep != null) {
                    o.add(dep);
                }
            }
        } finally {
            lock.unlock();
        }
        return new ServiceDependencyResolverImpl(build(m), build(o), serviceSets, serviceListings, properties);
    }
    
    protected abstract @Nullable Object register(final ServiceDependencyResolver resolver);
    protected final void registerService(final ServiceDependencyResolver resolver) {
        try {
            final Object serviceInstance = register(resolver);
            if (serviceInstance != null) {
                if (serviceInstance instanceof Reloadable) {
                    if (LOG.isTraceEnabled()) {
                        LOG.trace("'{}' is {}", serviceInstanceName, Reloadable.class.getSimpleName());
                    }
                    RegistrationTools.registerIntoContext(
                        context,
                        NullUtil.immutableSet(Reloadable.class),
                        serviceInstance,
                        null
                    );
                }
            }
        } finally {
        }
    }
    

    @Override
    public final void modifiedService(@Nullable ServiceReference<S> reference, @Nullable T service) {
        // Nothing to do.
    }

    @Override
    public final void removedService(@Nullable ServiceReference<S> reference, @Nullable T service) {
        if (reference == null || service == null) {
            return;
        }
        
        boolean someServiceMissing = false;
        lock.lock();
        try {
            for (final Map.Entry<MandatoryServiceDependency<?>, AtomicReference<ResolvedService>> e : mandatoryDependencyMap.entrySet()) {
                final ServiceDependency<?> dep = e.getKey();
                if (dep.matchesService(service, reference)) {
                    final ResolvedService oldService = e.getValue().getAndSet(null);
                    if (oldService != null) {
                        someServiceMissing = true;
                    }
                }
            }
            for (final OptionalServiceDependency<?> dep : optionalDependencyList) {
                if (dep.matchesService(service, reference)) {
                    dep.onRemoved();
                }
            }
        } finally {
            lock.unlock();
        }
        if (someServiceMissing) {
            unregister(service);
        }
        setState();
        context.ungetService(reference);
    }

    /**
     * You can overwrite this method to do something else to shut down your code if one of the needed services is gone. If you overwrote
     * {@link #register()} and you did not register an OSGi service, just ignore the unregister parameter. If you registered a servlet and
     * the HttpService is gone, it is not available anymore in neededServices but from the given service parameter.
     * @param unregister OSGi service registration that needs to be unregistered.
     * @param service OSGi service that is taken down.
     */
    protected abstract void unregister(Object service);

    private void setState() {
        final ServiceStateLookup lookup = DeferredActivator.getLookup();
        final List<String> missing = new ArrayList<>();
        final List<String> present = new ArrayList<>();
        lock.lock();
        try {
            for (final Map.Entry<MandatoryServiceDependency<?>, AtomicReference<ResolvedService>> e : mandatoryDependencyMap.entrySet()) {
                final String id = e.getKey().identify();
                if (e.getValue().get() == null) {
                    missing.add(id);
                } else {
                    present.add(id);
                }
            }
            for (final OptionalServiceDependency<?> dep : optionalDependencyList) {
                if (dep.serviceRef().isPresent()) {
                    present.add(dep.identify());
                }
                // don't add missing ones, as they are optional
            }
            lookup.setState(serviceInstanceName, missing, present);
        } finally {
            lock.unlock();
        }
    }
}
