/*
 * @copyright Copyright (c) OX Software GmbH, Germany <info@open-xchange.com>
 * @license AGPL-3.0
 *
 * This code is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with OX App Suite.  If not, see <https://www.gnu.org/licenses/agpl-3.0.txt>.
 *
 * Any use of the work other than as authorized under this license or copyright law is prohibited.
 *
 */

package com.openexchange.util.custom.base;

import static com.openexchange.util.custom.base.NullUtil.build;
import static com.openexchange.util.custom.base.NullUtil.notNull;
import static com.openexchange.util.custom.base.NullUtil.setBuilder;
import java.util.Map;
import java.util.function.Function;


import org.apache.commons.text.StringSubstitutor;
import org.apache.commons.text.lookup.StringLookup;
import com.google.common.collect.ImmutableSet;
import com.openexchange.annotation.Nullable;
import com.openexchange.config.ConfigurationService;

/**
 * {@link StringInterpolator}
 *
 * @author <a href="mailto:pascal.bleser@open-xchange.com">Pascal Bleser</a>
 * @since v1.0.3
 */
public final class StringInterpolator implements Function<String, String> {
    
    private static final class PlaceholderResolverTrackingStrLookup implements StringLookup {
        
        private final PlaceholderResolver resolver;
        private final ImmutableSet.Builder<String> interpolated = setBuilder();
        
        public PlaceholderResolverTrackingStrLookup(final PlaceholderResolver resolver) {
            super();
            this.resolver = resolver;
        }

        public ImmutableSet<String> interpolated() {
            return build(interpolated);
        }

        @Override
        public @Nullable String lookup(final @Nullable String key) {
            if (key == null) {
                return null;
            }
            
            final @Nullable String defaultValue;
            final String placeholder;
            {
                final int defaultMarkIndex = key.indexOf("|default=");
                if (defaultMarkIndex >= 0) {
                    placeholder = notNull(key.substring(0, defaultMarkIndex));
                    defaultValue = notNull(key.substring(defaultMarkIndex + "|default=".length()));
                    if (StringUtil.isEmptyOrWhitespaces(placeholder)) {
                        throw new FailedInterpolationException(key, "placeholder has a default value but the placeholder itself is empty");
                    }
                } else {
                    placeholder = key;
                    defaultValue = null;
                }
            }
            
            final String value;
            try {
                value = resolver.resolve(placeholder);
            } catch (final FailedInterpolationException e) {
                throw e;
            } catch (final Exception e) {
                throw new FailedInterpolationException(placeholder, e);
            }
            if (value == null) {
                if (defaultValue != null) {
                    interpolated.add(placeholder);                    
                    return defaultValue;
                } else {
                    throw new FailedInterpolationException(placeholder);
                }
            }
            interpolated.add(placeholder);
            return value;
        }
    }
    
    private enum NullPlaceholderResolver implements PlaceholderResolver {
        INSTANCE;
        @Override
        public @Nullable String resolve(String placeholder) throws Exception {
            return "";
        }
    }
    
    public static ImmutableSet<String> placeholders(final String template) {
        return new StringInterpolator(NullPlaceholderResolver.INSTANCE).interpolate(template).interpolated();
    }
    
    public static final StringInterpolator forConfigurationService(final ConfigurationService configurationService) {
        return new StringInterpolator(new ConfigurationServicePlaceholderResolver(configurationService));           
    }
    public static final StringInterpolator forMap(final Map<String, ?> map) {
        return new StringInterpolator(new MapPlaceholderResolver(map));
    }
    
    private final PlaceholderResolver resolver;
    
    public StringInterpolator(final PlaceholderResolver resolver) {
        this.resolver = resolver;
    }

    @Override
    @Nullable
    public String apply(@Nullable String input) throws FailedInterpolationException {
        if (input != null) {
            return interpolate(input).result();
        } else {
            return null;
        }
    }
    
    public StringInterpolation interpolate(final CharSequence input) throws FailedInterpolationException {
        return interpolate(input, "${", "}");
    }
    
    public StringInterpolation interpolate(final CharSequence input, final String prefix, final String suffix) throws FailedInterpolationException {
        final PlaceholderResolverTrackingStrLookup lookup = new PlaceholderResolverTrackingStrLookup(resolver);
        final String result = new StringSubstitutor(lookup, prefix, suffix, '\\').replace(input);
        if (result == null) {
            throw new IllegalStateException("interpolation result is null");
        }
        return new StringInterpolation(result, lookup.interpolated());
    }    
    
}
